/**
 * Copyright (c) 1993 Richard V. Nash.
 * Copyright (c) 2000 Dan Papasian.  All rights reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "includes.h"

static int player_zero(int p);

static int get_empty_slot(void)
{
  int i;

  for (i = 0; i < player_globals.p_num; i++) 
    if (player_globals.parray[i].status == PLAYER_EMPTY) 
      return i;
  
  player_globals.p_num++;
  
  if (player_globals.p_num+1 >= PARRAY_SIZE)
    fprintf(stderr, "*** Bogus attempt to get_empty_slot() past end of player_globals.parray ***\n");
  
  player_globals.parray[player_globals.p_num - 1].status = PLAYER_EMPTY;
  return player_globals.p_num - 1;
}

void player_array_init()
{
  int i;
  
  for (i = 0; i < PARRAY_SIZE; i++)
    player_globals.parray[i].status = PLAYER_EMPTY;
}

void player_init(int startConsole)
{
  int p;
  
  if (startConsole) {
    net_addConnection(0, 0);
    p = player_new();
    player_globals.parray[p].login = strdup("console");
    player_globals.parray[p].name = strdup("console");
    player_globals.parray[p].passwd = strdup("*");
    player_globals.parray[p].fullName = strdup("The Operator");
    player_globals.parray[p].emailAddress = NULL;
    player_globals.parray[p].prompt = strdup("chessd%");
    player_globals.parray[p].adminLevel = ADMIN_GOD;
    player_globals.parray[p].socket = 0;
    player_globals.parray[p].busy = NULL;
    pprintf_prompt(p, "\nLogged in on console.\n");
  }
}

int player_new()
{
  int new = get_empty_slot();
  
  player_zero(new);
  return new;
}

#define INVALID ((char *) -42)

static void ResetStats(statistics *ss)
{
  ss->num = 0;
  ss->win = 0;
  ss->los = 0;
  ss->dra = 0;
  ss->rating = 0;
  ss->sterr = 350.0;
  ss->ltime = 0;
  ss->best = 0;
  ss->whenbest = 0;
}

static int player_zero(int p)
{
  int i;

  player_globals.parray[p].name = NULL;
  player_globals.parray[p].login = NULL;
  player_globals.parray[p].fullName = NULL;
  player_globals.parray[p].emailAddress = NULL;
  player_globals.parray[p].prompt = config_get("DEFAULT_PROMPT");
  player_globals.parray[p].partner = -1;
  player_globals.parray[p].passwd = NULL;
  player_globals.parray[p].socket = -1;
  player_globals.parray[p].status = PLAYER_NEW;

  ResetStats(&player_globals.parray[p].s_stats);
  ResetStats(&player_globals.parray[p].b_stats);
  ResetStats(&player_globals.parray[p].l_stats);
  ResetStats(&player_globals.parray[p].w_stats);
  ResetStats(&player_globals.parray[p].bug_stats);

  player_globals.parray[p].d_time = 15;
  player_globals.parray[p].d_inc = 0;
  player_globals.parray[p].d_height = 24;
  player_globals.parray[p].d_width = 79;
  player_globals.parray[p].language = LANG_DEFAULT;
  player_globals.parray[p].last_file = NULL;
  player_globals.parray[p].last_file_byte = 0L;
  player_globals.parray[p].timeOfReg = 0;
  player_globals.parray[p].totalTime = 0;
  player_globals.parray[p].availmin = 0;
  player_globals.parray[p].availmax = 0;
  player_globals.parray[p].kiblevel = 0;
  player_globals.parray[p].style = 0;
  player_globals.parray[p].promote = QUEEN;
  player_globals.parray[p].game = -1;
  player_globals.parray[p].last_tell = -1;
  player_globals.parray[p].last_channel = -1;
  player_globals.parray[p].ftell = -1;
  player_globals.parray[p].logon_time = 0;
  player_globals.parray[p].last_command_time = 0;
  player_globals.parray[p].adminLevel = 0;
  player_globals.parray[p].Flags = PFLAG_DEFAULT;
  player_globals.parray[p].num_plan = 0;
  for (i = 0; i < MAX_PLAN; i++)
    player_globals.parray[p].planLines[i] = INVALID;
  player_globals.parray[p].num_formula = 0;
  for (i = 0; i < MAX_FORMULA; i++)
    player_globals.parray[p].formulaLines[i] = NULL;
  player_globals.parray[p].formula = NULL;
  player_globals.parray[p].num_white = 0;
  player_globals.parray[p].num_black = 0;
  player_globals.parray[p].num_observe = 0;
  player_globals.parray[p].thisHost = 0;
  player_globals.parray[p].lastHost = 0;
  player_globals.parray[p].number_pend_to = 0;
  player_globals.parray[p].number_pend_from = 0;
  player_globals.parray[p].numAlias = 0;
  for (i = 0; i < MAX_ALIASES; i++) {
    player_globals.parray[p].alias_list[i].comm_name = INVALID;
    player_globals.parray[p].alias_list[i].alias = INVALID;
  }
  player_globals.parray[p].opponent = -1;
  player_globals.parray[p].last_opponent = -1;
  player_globals.parray[p].highlight = 0;
  player_globals.parray[p].lastshout_a = 0;
  player_globals.parray[p].lastshout_b = 0;
  player_globals.parray[p].simul_info = NULL;
  player_globals.parray[p].num_comments = 0;
  player_globals.parray[p].lists = NULL;
  player_globals.parray[p].busy = NULL;
  return 0;
}

static int player_free(int p)
{
  int i;

  free(player_globals.parray[p].login);
  free(player_globals.parray[p].name);
  free(player_globals.parray[p].passwd);
  free(player_globals.parray[p].fullName);
  free(player_globals.parray[p].emailAddress);
  free(player_globals.parray[p].prompt);
  for (i = 0; i < player_globals.parray[p].num_plan; i++)
    free(player_globals.parray[p].planLines[i]);
  for (i = 0; i < player_globals.parray[p].num_formula; i++)
    free(player_globals.parray[p].formulaLines[i]);
  free(player_globals.parray[p].formula);
  list_free(player_globals.parray[p].lists);
  for (i = 0; i < player_globals.parray[p].numAlias; i++) {
    free(player_globals.parray[p].alias_list[i].comm_name);
    free(player_globals.parray[p].alias_list[i].alias);
  }
  if (player_globals.parray[p].busy != NULL)
    free (player_globals.parray[p].busy);
  if (player_globals.parray[p].simul_info != NULL)
    free (player_globals.parray[p].simul_info);
  return 0;
}

int player_clear(int p)
{
  player_free(p);
  player_zero(p);
  return 0;
}

int player_remove(int p)
{
  int i;

  decline_withdraw_offers(p, -1, -1, DO_DECLINE | DO_WITHDRAW);
  if (player_globals.parray[p].simul_info != NULL)
    /* Player disconnected in middle of simul */
    for (i = 0; player_globals.parray[p].simul_info != NULL
	   && i < player_globals.parray[p].simul_info->numBoards; i++)
      if (player_globals.parray[p].simul_info->boards[i] >= 0) 
	game_disconnect(player_globals.parray[p].simul_info->boards[i], p);
    
  if ((player_globals.parray[p].game >=0) && (pIsPlaying(p))) {  
    /* Player disconnected in the middle of a game! */
    pprintf(player_globals.parray[p].opponent, "Your opponent has lost contact or quit.");
    game_disconnect(player_globals.parray[p].game, p);
  }
  for (i = 0; i < player_globals.p_num; i++) {
    if (player_globals.parray[i].status == PLAYER_EMPTY)
      continue;
    if (player_globals.parray[i].last_tell == p)
      player_globals.parray[i].last_tell = -1;
    if (player_globals.parray[i].last_opponent == p)
      player_globals.parray[i].last_opponent = -1;
    if (player_globals.parray[i].partner == p) {
      pprintf_prompt (i, "Your partner has disconnected.\n");
      decline_withdraw_offers(i, -1, PEND_BUGHOUSE, DO_DECLINE | DO_WITHDRAW);
      player_globals.parray[i].partner = -1;
    }
  }
  player_clear(p);
  player_globals.parray[p].status = PLAYER_EMPTY;

  return 0;
}

/* return 0 if corrupt */

static int ReadV1PlayerFmt(int p,player *pp, FILE * fp, char *file, int version)
{

 int i,size_cens, size_noplay, size_not, size_gnot, size_chan, len;
 int bs,ss,ws,ls,bugs;
 int bell, pgn, notifiedby, pin, gin, avail, private, jprivate;
 int open, ropen, rated, tell, kib, shout, cshout, automail, mailmess;

 char* tmp;
 char tmp2[MAX_STRING_LENGTH];

 fgets(tmp2, MAX_STRING_LENGTH, fp);
 if (strcmp(tmp2,"NONE\n")) {
   tmp2[strlen(tmp2)-1] = '\0';
   pp->name = strdup (tmp2);
 } else
   pp->name = NULL;
 fgets(tmp2, MAX_STRING_LENGTH, fp);
 if (strcmp(tmp2,"NONE\n")) {
   tmp2[strlen(tmp2)-1] = '\0';
   pp->fullName = strdup (tmp2);
 } else
   pp->fullName = NULL;
 fgets(tmp2, MAX_STRING_LENGTH, fp);
 if (strcmp(tmp2,"NONE\n")) {
   tmp2[strlen(tmp2)-1] = '\0';
   pp->passwd = strdup (tmp2);
 } else
   pp->passwd = NULL;
 fgets(tmp2, MAX_STRING_LENGTH, fp);
 if (strcmp(tmp2,"NONE\n")) {
   tmp2[strlen(tmp2)-1] = '\0';
   pp->emailAddress = strdup (tmp2);
 } else
   pp->emailAddress = NULL;
 if (fscanf(fp, "%u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %d\n",

 &pp->s_stats.num, &pp->s_stats.win, &pp->s_stats.los,
 &pp->s_stats.dra, &pp->s_stats.rating, &ss,
 &pp->s_stats.ltime, &pp->s_stats.best, &pp->s_stats.whenbest,

 &pp->b_stats.num, &pp->b_stats.win, &pp->b_stats.los,
 &pp->b_stats.dra, &pp->b_stats.rating, &bs,
 &pp->b_stats.ltime, &pp->b_stats.best, &pp->b_stats.whenbest,

 &pp->w_stats.num, &pp->w_stats.win, &pp->w_stats.los,
 &pp->w_stats.dra, &pp->w_stats.rating, &ws,
 &pp->w_stats.ltime, &pp->w_stats.best, &pp->w_stats.whenbest,

 &pp->l_stats.num, &pp->l_stats.win, &pp->l_stats.los,
 &pp->l_stats.dra, &pp->l_stats.rating, &ls,
 &pp->l_stats.ltime, &pp->l_stats.best, &pp->l_stats.whenbest,

 &pp->bug_stats.num, &pp->bug_stats.win, &pp->bug_stats.los,
 &pp->bug_stats.dra, &pp->bug_stats.rating, &bugs,
 &pp->bug_stats.ltime, &pp->bug_stats.best, &pp->bug_stats.whenbest,
 &pp->lastHost) != 46) {
  fprintf(stderr,"Player %s is corrupt\n",player_globals.parray[p].name);
    free(pp->fullName);
    free(pp->passwd);
    free(pp->emailAddress);
    pp->fullName = NULL;
    pp->passwd = NULL;
    pp->emailAddress = NULL;
  return 0; 
 }

 pp->b_stats.sterr = bs / 10.0;
 pp->s_stats.sterr = ss / 10.0;
 pp->w_stats.sterr = ws / 10.0;
 pp->l_stats.sterr = ls / 10.0;
 pp->bug_stats.sterr = bugs / 10.0;

 fgets (tmp2, MAX_STRING_LENGTH, fp);
   tmp2[strlen(tmp2)-1] = '\0';
   pp->prompt = strdup(tmp2);
 if (version == 1) {
   if (fscanf (fp, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
   &open, &rated, &ropen, &pp->timeOfReg,
   &pp->totalTime, &bell, &pgn, &notifiedby,
   &pin, &gin, &shout, &cshout,
   &tell, &kib, &private, &jprivate,
   &automail, &mailmess, &pp->style, &pp->d_time,
   &pp->d_inc, &pp->d_height, &pp->d_width, &pp->language,
   &pp->adminLevel, &pp->num_white, &pp->num_black, &pp->highlight,
   &pp->num_comments,
   &pp->num_plan, &pp->num_formula,&size_cens,
   &size_not, &size_noplay,
   &size_gnot, &pp->numAlias, &size_chan) != 37) {
    fprintf(stderr,"Player %s is corrupt.\n",player_globals.parray[p].name);
    free(pp->prompt);
    pp->prompt = config_get("DEFAULT_PROMPT");
    free(pp->fullName);
    free(pp->passwd);
    free(pp->emailAddress);
    pp->fullName = NULL;
    pp->passwd = NULL;
    pp->emailAddress = NULL;
    pp->numAlias = 0;
    pp->num_plan = 0;
    pp->num_formula = 0;
    pp->adminLevel = 0;
    return 0;
   }
 } else if (version == 2) {
   if (fscanf (fp, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
   &open, &rated, &ropen, &pp->timeOfReg,
   &pp->totalTime, &bell, &pgn, &notifiedby,
   &pin, &gin, &avail, &shout, &cshout,
   &tell, &kib, &private, &jprivate,
   &automail, &mailmess, &pp->style, &pp->d_time,
   &pp->d_inc, &pp->d_height, &pp->d_width, &pp->language,
   &pp->adminLevel, &pp->num_white, &pp->num_black, &pp->highlight,
   &pp->num_comments,
   &pp->num_plan, &pp->num_formula,&size_cens,
   &size_not, &size_noplay,
   &size_gnot, &pp->numAlias, &size_chan) != 38) {
    fprintf(stderr,"Player %s is corrupt.\n",player_globals.parray[p].name);
    free(pp->prompt);
    pp->prompt = config_get("DEFAULT_PROMPT");
    free(pp->fullName);
    free(pp->passwd);
    free(pp->emailAddress);
    pp->fullName = NULL;
    pp->passwd = NULL;
    pp->emailAddress = NULL;
    pp->numAlias = 0;
    pp->num_plan = 0;
    pp->num_formula = 0;
    pp->adminLevel = 0;
    return 0;
   }
 } else if (version > 2) {
   if (fscanf (fp, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
   &open, &rated, &ropen, &pp->timeOfReg,
   &pp->totalTime, &bell, &pgn, &notifiedby,
   &pin, &gin, &avail, &shout, &cshout,
   &tell, &kib, &private, &jprivate,
   &automail, &mailmess, &pp->style, &pp->d_time,
   &pp->d_inc, &pp->d_height, &pp->d_width, &pp->language,
   &pp->adminLevel, &pp->num_white, &pp->num_black, &pp->highlight,
   &pp->num_comments,
   &pp->num_plan, &pp->num_formula,&size_cens,
   &size_not, &size_noplay,
   &size_gnot, &pp->numAlias, &size_chan, &pp->availmin, &pp->availmax) != 40) {
    fprintf(stderr,"Player %s is corrupt.\n",player_globals.parray[p].name);
    free(pp->prompt);
    pp->prompt = config_get("DEFAULT_PROMPT");
    free(pp->fullName);
    free(pp->passwd);
    free(pp->emailAddress);
    pp->fullName = NULL;
    pp->passwd = NULL;
    pp->emailAddress = NULL;
    pp->numAlias = 0;
    pp->num_plan = 0;
    pp->num_formula = 0;
    pp->adminLevel = 0;
    return 0;
   }
 }
 SetFlag(pp->Flags, PFLAG_OPEN, open);
 SetFlag(pp->Flags, PFLAG_ROPEN, ropen);
 SetFlag(pp->Flags, PFLAG_RATED, rated);
 SetFlag(pp->Flags, PFLAG_BELL, bell);
 SetFlag(pp->Flags, PFLAG_PGN, pgn);
 SetFlag(pp->Flags, PFLAG_NOTIFYBY, notifiedby);
 SetFlag(pp->Flags, PFLAG_PIN, pin);
 SetFlag(pp->Flags, PFLAG_GIN, gin);
 SetFlag(pp->Flags, PFLAG_AVAIL, avail);
 SetFlag(pp->Flags, PFLAG_PRIVATE, private);
 SetFlag(pp->Flags, PFLAG_JPRIVATE, jprivate);
 SetFlag(pp->Flags, PFLAG_SHOUT, shout);
 SetFlag(pp->Flags, PFLAG_CSHOUT, cshout);
 SetFlag(pp->Flags, PFLAG_TELL, tell);
 SetFlag(pp->Flags, PFLAG_KIBITZ, kib);
 SetFlag(pp->Flags, PFLAG_AUTOMAIL, automail);
 SetFlag(pp->Flags, PFLAG_MAILMESS, mailmess);

 if (version > 3)
   getc (fp); /* skip a dot */

   if (pp->num_plan > 0) {
     for (i = 0; i < pp->num_plan; i++) {
       fgets(tmp2, MAX_LINE_SIZE, fp);
       if (!(len = strlen(tmp2))) {
         fprintf(stderr, "CHESSD: Error bad plan in file %s\n", file);
         i--;
         pp->num_plan--; 
       } else {
         tmp2[len - 1] = '\0';  /* Get rid of '\n' */
         pp->planLines[i] = (len > 1) ? strdup(tmp2) : NULL;
       }
     }
   }
   if (pp->num_formula > 0) {
     for (i = 0; i < pp->num_formula; i++) {
       fgets(tmp2, MAX_LINE_SIZE, fp);
       if (!(len = strlen(tmp2))) {
         fprintf(stderr, "CHESSD: Error bad formula in file %s\n", file);
         i--;
         pp->num_formula--;
       } else {
         tmp2[len - 1] = '\0';  /* Get rid of '\n' */
         pp->formulaLines[i] = (len > 1) ? strdup(tmp2) : NULL;
       }
     }
   }
  fgets(tmp2, MAX_LINE_SIZE, fp);
  tmp2[strlen(tmp2) - 1] = '\0';
  if (!strcmp (tmp2,"NONE"))
    pp->formula = NULL;
  else
    pp->formula = strdup(tmp2);

  if (pp->numAlias > 0) {
    for (i = 0; i < pp->numAlias; i++) {
      fgets(tmp2, MAX_LINE_SIZE, fp);
      if (!(len = strlen(tmp2))) {
        fprintf(stderr, "CHESSD: Error bad alias in file %s\n", file);
        i--;
        pp->numAlias--;
      } else {
        tmp2[len - 1] = '\0';  /* Get rid of '\n' */
        tmp = tmp2;
        tmp = eatword(tmp2);
        *tmp = '\0';
        tmp++;
        tmp = eatwhite(tmp);
        pp->alias_list[i].comm_name = strdup(tmp2);
        pp->alias_list[i].alias = strdup(tmp);
      }
    }
  }

    while (size_cens--) {
      fscanf(fp,"%s",tmp2);
      list_add(p, L_CENSOR, tmp2);
      }

    while(size_not--) {
      fscanf(fp,"%s",tmp2);
      list_add(p, L_NOTIFY, tmp2);
      }

    while(size_noplay--) {
      fscanf(fp,"%s",tmp2);
      list_add(p, L_NOPLAY, tmp2);
      }

    while(size_gnot--) {
      fscanf(fp,"%s",tmp2);
      list_add(p, L_GNOTIFY, tmp2);
      }

    while(size_chan--) {
      fscanf(fp,"%s",tmp2);
      list_add(p, L_CHANNEL, tmp2);
      }
  return 1;
}

static int got_attr_value_player(int p, char *attr, char *value, FILE * fp, char *file)
{
  int i, len;
  char tmp[MAX_LINE_SIZE], *tmp1;

  if (!strcmp(attr, "name:")) {
    player_globals.parray[p].name = strdup(value);
  } else if (!strcmp(attr, "password:")) {
    player_globals.parray[p].passwd = strdup(value);
  } else if (!strcmp(attr, "fullname:")) {
    player_globals.parray[p].fullName = strdup(value);
  } else if (!strcmp(attr, "email:")) {
    player_globals.parray[p].emailAddress = strdup(value);
  } else if (!strcmp(attr, "prompt:")) {
    player_globals.parray[p].prompt = strdup(value);
  } else if (!strcmp(attr, "s_num:")) {
    player_globals.parray[p].s_stats.num = atoi(value);
  } else if (!strcmp(attr, "s_win:")) {
    player_globals.parray[p].s_stats.win = atoi(value);
  } else if (!strcmp(attr, "s_loss:")) {
    player_globals.parray[p].s_stats.los = atoi(value);
  } else if (!strcmp(attr, "s_draw:")) {
    player_globals.parray[p].s_stats.dra = atoi(value);
  } else if (!strcmp(attr, "s_rating:")) {
    player_globals.parray[p].s_stats.rating = atoi(value);
  } else if (!strcmp(attr, "s_sterr:")) {
    player_globals.parray[p].s_stats.sterr = (atoi(value) / 10.0);
  } else if (!strcmp(attr, "s_ltime:")) {
    player_globals.parray[p].s_stats.ltime = atoi(value);
  } else if (!strcmp(attr, "s_best:")) {
    player_globals.parray[p].s_stats.best = atoi(value);
  } else if (!strcmp(attr, "s_wbest:")) {
    player_globals.parray[p].s_stats.whenbest = atoi(value);
  } else if (!strcmp(attr, "b_num:")) {
    player_globals.parray[p].b_stats.num = atoi(value);
  } else if (!strcmp(attr, "b_win:")) {
    player_globals.parray[p].b_stats.win = atoi(value);
  } else if (!strcmp(attr, "b_loss:")) {
    player_globals.parray[p].b_stats.los = atoi(value);
  } else if (!strcmp(attr, "b_draw:")) {
    player_globals.parray[p].b_stats.dra = atoi(value);
  } else if (!strcmp(attr, "b_rating:")) {
    player_globals.parray[p].b_stats.rating = atoi(value);
  } else if (!strcmp(attr, "b_sterr:")) {
    player_globals.parray[p].b_stats.sterr = (atoi(value) / 10.0);
  } else if (!strcmp(attr, "b_ltime:")) {
    player_globals.parray[p].b_stats.ltime = atoi(value);
  } else if (!strcmp(attr, "b_best:")) {
    player_globals.parray[p].b_stats.best = atoi(value);
  } else if (!strcmp(attr, "b_wbest:")) {
    player_globals.parray[p].b_stats.whenbest = atoi(value);
  } else if (!strcmp(attr, "w_num:")) {
    player_globals.parray[p].w_stats.num = atoi(value);
  } else if (!strcmp(attr, "w_win:")) {
    player_globals.parray[p].w_stats.win = atoi(value);
  } else if (!strcmp(attr, "w_loss:")) {
    player_globals.parray[p].w_stats.los = atoi(value);
  } else if (!strcmp(attr, "w_draw:")) {
    player_globals.parray[p].w_stats.dra = atoi(value);
  } else if (!strcmp(attr, "w_rating:")) {
    player_globals.parray[p].w_stats.rating = atoi(value);
  } else if (!strcmp(attr, "w_sterr:")) {
    player_globals.parray[p].w_stats.sterr = (atoi(value) / 10.0);
  } else if (!strcmp(attr, "w_ltime:")) {
    player_globals.parray[p].w_stats.ltime = atoi(value);
  } else if (!strcmp(attr, "w_best:")) {
    player_globals.parray[p].w_stats.best = atoi(value);
  } else if (!strcmp(attr, "w_wbest:")) {
    player_globals.parray[p].w_stats.whenbest = atoi(value);
  } else if (!strcmp(attr, "open:")) {
    SetPFlag(p, PFLAG_OPEN, atoi(value));
  } else if (!strcmp(attr, "rated:")) {
    SetPFlag(p, PFLAG_RATED, atoi(value));
  } else if (!strcmp(attr, "ropen:")) {
    SetPFlag(p, PFLAG_ROPEN, atoi(value));
  } else if (!strcmp(attr, "bell:")) {
    SetPFlag(p, PFLAG_BELL, atoi(value));
  } else if (!strcmp(attr, "pgn:")) {
    SetPFlag(p, PFLAG_PGN, atoi(value));
  } else if (!strcmp(attr, "timeofreg:")) {
    player_globals.parray[p].timeOfReg = atoi(value);
  } else if (!strcmp(attr, "totaltime:")) {
    player_globals.parray[p].totalTime = atoi(value);
  } else if (!strcmp(attr, "notifiedby:")) {
    SetPFlag(p, PFLAG_NOTIFYBY, atoi(value));
  } else if (!strcmp(attr, "i_login:")) {
    SetPFlag(p, PFLAG_PIN, atoi(value));
  } else if (!strcmp(attr, "i_game:")) {
    SetPFlag(p, PFLAG_GIN, atoi(value));
  } else if (!strcmp(attr, "i_shout:")) {
    SetPFlag(p, PFLAG_SHOUT, atoi(value));
  } else if (!strcmp(attr, "i_cshout:")) {
    SetPFlag(p, PFLAG_CSHOUT, atoi(value));
  } else if (!strcmp(attr, "i_tell:")) {
    SetPFlag(p, PFLAG_TELL, atoi(value));
  } else if (!strcmp(attr, "i_kibitz:")) {
    SetPFlag(p, PFLAG_KIBITZ, atoi(value));
  } else if (!strcmp(attr, "kiblevel:")) {
    player_globals.parray[p].kiblevel = atoi(value);
  } else if (!strcmp(attr, "private:")) {
    SetPFlag(p, PFLAG_PRIVATE, atoi(value));
  } else if (!strcmp(attr, "jprivate:")) {
    SetPFlag(p, PFLAG_JPRIVATE, atoi(value));
  } else if (!strcmp(attr, "automail:")) {
    SetPFlag(p, PFLAG_AUTOMAIL, atoi(value));
  } else if (!strcmp(attr, "i_mailmess:")) {
    SetPFlag(p, PFLAG_MAILMESS, atoi(value));
  } else if (!strcmp(attr, "style:")) {
    player_globals.parray[p].style = atoi(value);
  } else if (!strcmp(attr, "d_time:")) {
    player_globals.parray[p].d_time = atoi(value);
  } else if (!strcmp(attr, "d_inc:")) {
    player_globals.parray[p].d_inc = atoi(value);
  } else if (!strcmp(attr, "d_height:")) {
    player_globals.parray[p].d_height = atoi(value);
  } else if (!strcmp(attr, "d_width:")) {
    player_globals.parray[p].d_width = atoi(value);
  } else if (!strcmp(attr, "language:")) {
    player_globals.parray[p].language = atoi(value);
  } else if (!strcmp(attr, "admin_level:")) {
    player_globals.parray[p].adminLevel = atoi(value);
    if (player_globals.parray[p].adminLevel >= ADMIN_ADMIN)
      PFlagON(p, PFLAG_ADMINLIGHT);
  } else if (!strcmp(attr, "i_admin:")) {
/*    player_globals.parray[p].i_admin = atoi(value);  */
  } else if (!strcmp(attr, "computer:")) {
/*    player_globals.parray[p].computer = atoi(value); */
  } else if (!strcmp(attr, "black_games:")) {
    player_globals.parray[p].num_black = atoi(value);
  } else if (!strcmp(attr, "white_games:")) {
    player_globals.parray[p].num_white = atoi(value);
  } else if (!strcmp(attr, "uscf:")) {
/*    player_globals.parray[p].uscfRating = atoi(value); */
  } else if (!strcmp(attr, "muzzled:")) {	/* ignore these: obsolete */
  } else if (!strcmp(attr, "cmuzzled:")) {	/* ignore these: obsolete */
  } else if (!strcmp(attr, "highlight:")) {
    player_globals.parray[p].highlight = atoi(value);
  } else if (!strcmp(attr, "network:")) {
/*    player_globals.parray[p].network_player = atoi(value); */
  } else if (!strcmp(attr, "lasthost:")) {
    player_globals.parray[p].lastHost = atoi(value);
  } else if (!strcmp(attr, "channel:")) {
    list_addsub(p,"channel",value, 1);
  } else if (!strcmp(attr, "num_comments:")) {
    player_globals.parray[p].num_comments = atoi(value);
  } else if (!strcmp(attr, "num_plan:")) {
    player_globals.parray[p].num_plan = atoi(value);
    if (player_globals.parray[p].num_plan > 0) {
      for (i = 0; i < player_globals.parray[p].num_plan; i++) {
	fgets(tmp, MAX_LINE_SIZE, fp);
	if (!(len = strlen(tmp))) {
	  fprintf(stderr, "CHESSD: Error bad plan in file %s\n", file);
          i--;
          player_globals.parray[p].num_plan--;
	} else {
	  tmp[len - 1] = '\0';	/* Get rid of '\n' */
	  player_globals.parray[p].planLines[i] = (len > 1) ? strdup(tmp) : NULL;
	}
      }
    }
  } else if (!strcmp(attr, "num_formula:")) {
    player_globals.parray[p].num_formula = atoi(value);
    if (player_globals.parray[p].num_formula > 0) {
      for (i = 0; i < player_globals.parray[p].num_formula; i++) {
	fgets(tmp, MAX_LINE_SIZE, fp);
	if (!(len = strlen(tmp))) {
	  fprintf(stderr, "CHESSD: Error bad formula in file %s\n", file);
          i--;
          player_globals.parray[p].num_formula--;
	} else {
	  tmp[len - 1] = '\0';	/* Get rid of '\n' */
	  player_globals.parray[p].formulaLines[i] = (len > 1) ? strdup(tmp) : NULL;
	}
      }
    }
  } else if (!strcmp(attr, "formula:")) {
    player_globals.parray[p].formula = strdup(value);
  } else if (!strcmp(attr, "num_alias:")) {
    player_globals.parray[p].numAlias = atoi(value);
    if (player_globals.parray[p].numAlias > 0) {
      for (i = 0; i < player_globals.parray[p].numAlias; i++) {
	fgets(tmp, MAX_LINE_SIZE, fp);
	if (!(len = strlen(tmp))) {
	  fprintf(stderr, "CHESSD: Error bad alias in file %s\n", file);
	  i--;
	  player_globals.parray[p].numAlias--;
	} else {
	  tmp[len - 1] = '\0';	/* Get rid of '\n' */
	  tmp1 = tmp;
	  tmp1 = eatword(tmp1);
	  *tmp1 = '\0';
	  tmp1++;
	  tmp1 = eatwhite(tmp1);
	  player_globals.parray[p].alias_list[i].comm_name = strdup(tmp);
	  player_globals.parray[p].alias_list[i].alias = strdup(tmp1);
	}
      }
    }
  } else if (!strcmp(attr, "num_censor:")) {
    i = atoi(value);
    while (i--) {
      fgets(tmp, MAX_LINE_SIZE, fp);
      if ((!(len = strlen(tmp))) || (len == 1)) { /* blank lines do occur!! */
        fprintf(stderr, "CHESSD: Error bad censor in file %s\n", file);
      } else {
        tmp[len - 1] = '\0';    /* Get rid of '\n' */
        list_add(p, L_CENSOR, tmp);
      }
    }
  } else if (!strcmp(attr, "num_notify:")) {
    i = atoi(value);
    while(i--) {
      fgets(tmp, MAX_LINE_SIZE, fp);
      if ((!(len = strlen(tmp))) || (len == 1)) { /* blank lines do occur!! */
        fprintf(stderr, "CHESSD: Error bad notify in file %s\n", file);
      } else {
        tmp[len - 1] = '\0';    /* Get rid of '\n' */
        list_add(p, L_NOTIFY, tmp);
      }
    }
  } else if (!strcmp(attr, "num_noplay:")) {
    i = atoi(value);
    while(i--) {
      fgets(tmp, MAX_LINE_SIZE, fp);
      if ((!(len = strlen(tmp))) || (len == 1)) { /* blank lines do occur!! */
        fprintf(stderr, "CHESSD: Error bad noplay in file %s\n", file);
      } else {
        tmp[len - 1] = '\0';    /* Get rid of '\n' */
        list_add(p, L_NOPLAY, tmp);
      }
    }
  } else if (!strcmp(attr, "num_gnotify:")) {
    i = atoi(value);
    while(i--) {
      fgets(tmp, MAX_LINE_SIZE, fp);
      if ((!(len = strlen(tmp)))  || (len == 1)) { /* blank lines do occur!! */
        fprintf(stderr, "CHESSD: Error bad gnotify in file %s\n", file);
      } else {
        tmp[len - 1] = '\0';    /* Get rid of '\n' */
        list_add(p, L_GNOTIFY, tmp);
      }
    }
  } else {
    fprintf(stderr, "CHESSD: Error bad attribute >%s< from file %s\n", attr, file);
  }
  return 0;
}

int player_read(int p, char *name)
{
  char fname[MAX_FILENAME_SIZE];
  char line[MAX_LINE_SIZE];
  char *attr, *value;
  FILE *fp;
  int len;
  int version = 0;

  player_globals.parray[p].login = stolower(strdup(name));

  sprintf(fname, "%s/%c/%s", PLAYER_DIR, player_globals.parray[p].login[0], player_globals.parray[p].login);
  fp = fopen(fname, "r");

  if (player_globals.parray[p].name)
	  free(player_globals.parray[p].name);

  if (!fp) { /* unregistered player */
    player_globals.parray[p].name = stolower(strdup(name));
    PFlagOFF(p, PFLAG_REG);
    return -1;
  }

  PFlagON(p, PFLAG_REG); /* lets load the file */

  fgets(line, MAX_LINE_SIZE, fp); /* ok so which version file? */

  if (line[0] == 'v') {
    sscanf(line, "%*c %d", &version);
  }
  if (version > PLAYER_VERSION) { /* eg if two version of code running at diff sites */
    pprintf(p, "Your account's version number is newer than that of the version running.\n"); 
    pprintf(p, "Please see an admin.\n");
    fprintf (stderr,"Player %s's version number is incorrect for this version.\n",player_globals.parray[p].name);
    PFlagOFF(p, PFLAG_REG);
    player_globals.parray[p].name = strdup(name);
    fclose (fp);
    return -1;
  }
  if (version > 0) {
    if (!ReadV1PlayerFmt(p,&player_globals.parray[p], fp, fname, version)) {/* Quick method */
      /* error in reading; make unreg */
      PFlagOFF(p, PFLAG_REG);
      PFlagOFF(p, PFLAG_RATED);
      pprintf(p, "\n*** WARNING: Your Data file is corrupt. Please tell an admin ***\n");
      fclose (fp);
      return -1; /* other data has already been freed */
    }
  }
  else /* do it the old SLOW way */
   do {
    if (feof(fp))
      break;
    if ((len = strlen(line)) <= 1)
      continue;
    line[len - 1] = '\0';
    attr = eatwhite(line);
    if (attr[0] == '#')
      continue;			/* Comment */
    value = eatword(attr);
    if (!*value) {
      fprintf(stderr, "CHESSD: Error reading file %s\n", fname);
      continue;
    }
    *value = '\0';
    value++;
    value = eatwhite(value);
    stolower(attr);
    got_attr_value_player(p, attr, value, fp, fname);
    fgets(line, MAX_LINE_SIZE, fp);
   } while (!feof(fp));

  fclose(fp);

  if (version == 0)
   player_save (p); /* ensure old files are quickly converted eg when someone
						fingers */
  if (!player_globals.parray[p].name) {
    player_globals.parray[p].name = strdup(name);
    pprintf(p, "\n*** WARNING: Your Data file is corrupt. Please tell an admin ***\n");
  }

  return 0;
}

static void WritePlayerFile(FILE* fp, int p)
{
 int i;
 player *pp = &player_globals.parray[p];

 fprintf (fp, "v %d\n", PLAYER_VERSION);
 if (pp->name == NULL) /* This should never happen! */
   fprintf (fp,"NONE\n");
 else
   fprintf(fp, "%s\n", pp->name);
 if (pp->fullName == NULL)
   fprintf (fp,"NONE\n");
 else
   fprintf(fp, "%s\n", pp->fullName);
 if (pp->passwd == NULL)
   fprintf (fp,"NONE\n");
 else
   fprintf(fp, "%s\n", pp->passwd);
 if (pp->emailAddress == NULL)
   fprintf (fp,"NONE\n");
 else
   fprintf(fp, "%s\n", pp->emailAddress); 
 fprintf(fp, "%u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %d\n",

 pp->s_stats.num, pp->s_stats.win, pp->s_stats.los, 
 pp->s_stats.dra, pp->s_stats.rating, (int) (pp->s_stats.sterr * 10.0),
 pp->s_stats.ltime, pp->s_stats.best, pp->s_stats.whenbest,

 pp->b_stats.num, pp->b_stats.win, pp->b_stats.los,
 pp->b_stats.dra, pp->b_stats.rating, (int) (pp->b_stats.sterr * 10.0),
 pp->b_stats.ltime, pp->b_stats.best, pp->b_stats.whenbest,

 pp->w_stats.num, pp->w_stats.win, pp->w_stats.los,
 pp->w_stats.dra, pp->w_stats.rating, (int) (pp->w_stats.sterr * 10.0),
 pp->w_stats.ltime, pp->w_stats.best, pp->w_stats.whenbest,

 pp->l_stats.num, pp->l_stats.win, pp->l_stats.los,
 pp->l_stats.dra, pp->l_stats.rating, (int) (pp->l_stats.sterr * 10.0),
 pp->l_stats.ltime, pp->l_stats.best, pp->l_stats.whenbest,

 pp->bug_stats.num, pp->bug_stats.win, pp->bug_stats.los,
 pp->bug_stats.dra, pp->bug_stats.rating, (int) (pp->bug_stats.sterr * 10.0),
 pp->bug_stats.ltime, pp->bug_stats.best, pp->bug_stats.whenbest,
 pp->lastHost);

 fprintf (fp, "%s\n", pp->prompt);
 fprintf (fp, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
 BoolCheckPFlag(p, PFLAG_OPEN), BoolCheckPFlag(p, PFLAG_RATED), 
 BoolCheckPFlag(p, PFLAG_ROPEN), pp->timeOfReg,
 pp->totalTime, BoolCheckPFlag(p, PFLAG_BELL),
 BoolCheckPFlag(p, PFLAG_PGN), BoolCheckPFlag(p, PFLAG_NOTIFYBY),
 BoolCheckPFlag(p, PFLAG_PIN), BoolCheckPFlag(p, PFLAG_GIN),
 BoolCheckPFlag(p, PFLAG_AVAIL), BoolCheckPFlag(p, PFLAG_SHOUT),
 BoolCheckPFlag(p, PFLAG_CSHOUT), BoolCheckPFlag(p, PFLAG_TELL),
 BoolCheckPFlag(p, PFLAG_KIBITZ), BoolCheckPFlag(p, PFLAG_PRIVATE),
 BoolCheckPFlag(p, PFLAG_JPRIVATE), BoolCheckPFlag(p, PFLAG_AUTOMAIL),
 BoolCheckPFlag(p, PFLAG_MAILMESS), pp->style,
 pp->d_time, pp->d_inc, pp->d_height, pp->d_width, pp->language,
 pp->adminLevel, pp->num_white, pp->num_black, pp->highlight,
 pp->num_comments,
 pp->num_plan, pp->num_formula,list_size(p, L_CENSOR),
 list_size(p, L_NOTIFY), list_size(p, L_NOPLAY),
 list_size(p, L_GNOTIFY), pp->numAlias, list_size(p, L_CHANNEL ), pp->availmin, pp->availmax);

  fprintf (fp,"."); /* Dot before plan - stop ill affects of scanf */

  for (i = 0; i < pp->num_plan; i++)
    fprintf(fp, "%s\n", (pp->planLines[i] ? pp->planLines[i] : ""));
  for (i = 0; i < pp->num_formula; i++)
    fprintf(fp, "%s\n", (pp->formulaLines[i] ? pp->formulaLines[i] : ""));
  if (player_globals.parray[p].formula != NULL)
    fprintf(fp, "%s\n", pp->formula);
  else
    fprintf(fp, "NONE\n");
  for (i = 0; i < pp->numAlias; i++)
    fprintf(fp, "%s %s\n", pp->alias_list[i].comm_name,
	    pp->alias_list[i].alias);

  list_print(fp, p, L_CENSOR);
  list_print(fp, p, L_NOTIFY);
  list_print(fp, p, L_NOPLAY);
  list_print(fp, p, L_GNOTIFY);
  list_print(fp, p, L_CHANNEL);
}

int player_save(int p)
{
  FILE *fp;
  char fname[MAX_FILENAME_SIZE];

  if (!CheckPFlag(p, PFLAG_REG)) { /* Player not registered */
    return -1;
  }
  if (player_globals.parray[p].name == NULL) { /* fixes a bug if name is null */
	    pprintf(p, "WARNING: Your player file could not be updated, due to corrupt data.\n");
    return -1;
   }
  if (strcasecmp(player_globals.parray[p].login, player_globals.parray[p].name)) {
    pprintf(p, "WARNING: Your player file could not be updated, due to corrupt data.\n");
    return -1;
  }

  sprintf(fname, "%s/%c/%s", PLAYER_DIR, player_globals.parray[p].login[0], player_globals.parray[p].login);
  fp = fopen(fname, "w");
  if (!fp) {
    fprintf(stderr, "CHESSD: Problem opening file %s for write\n", fname);
    return -1;
  }

  WritePlayerFile(fp,p);
  fclose(fp);
  return 0;
}

int player_find(int fd)
{
  int i;

  for (i = 0; i < player_globals.p_num; i++) {
    if (player_globals.parray[i].status == PLAYER_EMPTY)
      continue;
    if (player_globals.parray[i].socket == fd)
      return i;
  }
  return -1;
}

/* incorrectly named */
int player_find_bylogin(char *name)
{
  int i;

  for (i = 0; i < player_globals.p_num; i++) {
    if ((player_globals.parray[i].status == PLAYER_EMPTY) ||
	(player_globals.parray[i].status == PLAYER_LOGIN) ||
	(player_globals.parray[i].status == PLAYER_PASSWORD))
      continue;
    if (!player_globals.parray[i].login)
      continue;
    if (!strcasecmp(player_globals.parray[i].name, name))
      return i;
  }
  return -1;
}

/* Now incorrectly named */
int player_find_part_login(char *name)
{
  int i;
  int found = -1;

  i = player_find_bylogin(name);
  if (i >= 0)
    return i;
  for (i = 0; i < player_globals.p_num; i++) {
    if ((player_globals.parray[i].status == PLAYER_EMPTY) ||
	(player_globals.parray[i].status == PLAYER_LOGIN) ||
	(player_globals.parray[i].status == PLAYER_PASSWORD))
      continue;
    if (!player_globals.parray[i].name)
      continue;
    if (!strncasecmp(player_globals.parray[i].name, name, strlen(name))) {
      found = i;
      break;
    }
  }

  if(found < 0)
    return -2;
  else
    return found;
}

int player_censored(int p, int p1)
{
  if (in_list(p, L_CENSOR, player_globals.parray[p1].login))
    return 1;
  else
    return 0;
}

/* is p1 on p's notify list? */
int player_notified(int p, int p1)
{
  if (!CheckPFlag(p1, PFLAG_REG))
    return 0;

  /* possible bug: p has just arrived! */
  if (!player_globals.parray[p].name)
    return 0;

  return (in_list(p, L_NOTIFY, player_globals.parray[p1].login));
}

void player_notify_departure(int p)
/* Notify those with notifiedby set on a departure */
{
  int p1;

  if (!CheckPFlag(p, PFLAG_REG))
    return;
  for (p1 = 0; p1 < player_globals.p_num; p1++) {
    if (CheckPFlag(p1, PFLAG_NOTIFYBY) && !player_notified(p1, p)
        && player_notified(p, p1) && (player_globals.parray[p1].status == PLAYER_PROMPT)) {
      Bell (p1);
      pprintf(p1, "\nNotification: ");
      pprintf_highlight(p1, "%s", player_globals.parray[p].name);
      pprintf_prompt(p1, " has departed and isn't on your notify list.\n");
    }
  }
}

int player_notify_present(int p)
/* output Your arrival was notified by..... */
/* also notify those with notifiedby set if necessary */
{
  int p1;
  int count = 0;

  if (!CheckPFlag(p, PFLAG_REG))
    return count;
  for (p1 = 0; p1 < player_globals.p_num; p1++) {
    if ((player_notified(p, p1)) && (player_globals.parray[p1].status == PLAYER_PROMPT)) {
      if (!count) {
	pprintf(p, "Present company includes:");
      }
      count++;
      pprintf(p, " %s", player_globals.parray[p1].name);
      if (CheckPFlag(p1, PFLAG_NOTIFYBY) && (!player_notified(p1, p))
                         && (player_globals.parray[p1].status == PLAYER_PROMPT)) {
	Bell (p1);
	pprintf(p1, "\nNotification: ");
	pprintf_highlight(p1, "%s", player_globals.parray[p].name);
	pprintf_prompt(p1, " has arrived and isn't on your notify list.\n");
      }
    }
  }
  if (count)
    pprintf(p, ".\n");
  return count;
}

int player_notify(int p, char *note1, char *note2)
/* notify those interested that p has arrived/departed */
{
  int p1, count = 0;

  if (!CheckPFlag(p, PFLAG_REG))
    return count;
  for (p1 = 0; p1 < player_globals.p_num; p1++) {
    if ((player_notified(p1, p)) && (player_globals.parray[p1].status == PLAYER_PROMPT)) {
      Bell (p1);
      pprintf(p1, "\nNotification: ");
      pprintf_highlight(p1, "%s", player_globals.parray[p].name);
      pprintf_prompt(p1, " has %s.\n", note1);
      if (!count) {
	pprintf(p, "Your %s was noted by:", note2);
      }
      count++;
      pprintf(p, " %s", player_globals.parray[p1].name);
    }
  }
  if (count)
    pprintf(p, ".\n");
  return count;
}

int showstored(int p)
{
  DIR *dirp;
  struct dirent *dp;

  int c=0,p1;
  char dname[MAX_FILENAME_SIZE];
  multicol *m = multicol_start(50); /* Limit to 50, should be enough*/
  
  sprintf(dname, "%s/%c", ADJOURNED_DIR, player_globals.parray[p].login[0]);
  dirp = opendir(dname);
  if (!dirp) {
    multicol_end(m);
    return COM_OK;
  }
  for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
    if (file_has_pname(dp->d_name, player_globals.parray[p].login)) {
      if (strcmp(file_wplayer(dp->d_name),player_globals.parray[p].login) != 0) {
      	p1=player_find_bylogin(file_wplayer(dp->d_name));
      } else {
      	p1=player_find_bylogin(file_bplayer(dp->d_name));
      }
      if (p1>=0) {
      	if (c<50)
      		multicol_store(m,player_globals.parray[p1].name);
      	pprintf(p1,"\nNotification: ");
      	pprintf_highlight(p1,"%s",player_globals.parray[p].name);
      	pprintf_prompt(p1,", who has an adjourned game with you, has arrived.\n");
      	c++;
      } 
    }
  }
  closedir(dirp);
  if (c == 1) {
        pprintf(p, "1 player, who has an adjourned game with you, is online:\007");
  } else if (c > 1) {
  	pprintf(p, "\n%d players, who have an adjourned game with you, are online:\007",c);
  }
  if (c != 0)
  	multicol_pprint(m,p,player_globals.parray[p].d_width,2);
  multicol_end(m);

  return COM_OK;
}

int player_count(int CountAdmins)
{
  int count;
  int i;

  for (count = 0, i = 0; i < player_globals.p_num; i++) {
    if ((player_globals.parray[i].status == PLAYER_PROMPT) &&
        (CountAdmins || !in_list(i, L_ADMIN, player_globals.parray[i].name)))
      count++;
  }
  if (count > command_globals.player_high)
    command_globals.player_high = count;

  return count;
}

int player_idle(int p)
{
  if (player_globals.parray[p].status != PLAYER_PROMPT)
    return time(0) - player_globals.parray[p].logon_time;
  else
    return time(0) - player_globals.parray[p].last_command_time;
}

int player_ontime(int p)
{
  return time(0) - player_globals.parray[p].logon_time;
}

static void write_p_inout(int inout, int p, char *file, int maxlines)
{
  FILE *fp;

  fp = fopen(file, "a");
  if (!fp)
    return;
  fprintf(fp, "%d %s %d %d %s\n", inout, player_globals.parray[p].name, (int) time(0),
	  BoolCheckPFlag(p, PFLAG_REG),
          dotQuad(player_globals.parray[p].thisHost));
  fclose(fp);
  if (maxlines)
    truncate_file(file, maxlines);
}

void player_write_login(int p)
{
  char fname[MAX_FILENAME_SIZE];

  if (CheckPFlag(p, PFLAG_REG)) {
    sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR, player_globals.parray[p].login[0], player_globals.parray[p].login, STATS_LOGONS);
    write_p_inout(P_LOGIN, p, fname, 8);
  }
  sprintf(fname, "%s/%s", STATS_DIR, STATS_LOGONS);
  write_p_inout(P_LOGIN, p, fname, 30);
  /* added complete login/logout log to "logons.log" file */
  sprintf(fname, "%s/%s", STATS_DIR, "logons.log");
  write_p_inout(P_LOGIN, p, fname, 0);
}

void player_write_logout(int p)
{
  char fname[MAX_FILENAME_SIZE];

  if (CheckPFlag(p, PFLAG_REG)) {
    sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR, player_globals.parray[p].login[0], player_globals.parray[p].login, STATS_LOGONS);
    write_p_inout(P_LOGOUT, p, fname, 8);
  }
  sprintf(fname, "%s/%s", STATS_DIR, STATS_LOGONS);
  write_p_inout(P_LOGOUT, p, fname, 30);
  /* added complete login/logout log to "logons.log" file */
  sprintf(fname, "%s/%s", STATS_DIR, "logons.log");
  write_p_inout(P_LOGOUT, p, fname, 0);
}

int player_lastdisconnect(int p)
{
  char fname[MAX_FILENAME_SIZE];
  FILE *fp;
  int inout, thetime, registered;
  int last = 0;
  char ipstr[20];
  char loginName[MAX_LOGIN_NAME];

  sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR, player_globals.parray[p].login[0], player_globals.parray[p].login, STATS_LOGONS);
  fp = fopen(fname, "r");
  if (!fp)
    return 0;
  while (!feof(fp)) {
    if (fscanf(fp, "%d %s %d %d %s\n", &inout, loginName, &thetime, &registered, ipstr) != 5) {
      fprintf(stderr, "CHESSD: Error in login info format. %s\n", fname);
      fclose(fp);
      return 0;
    }
    if (inout == P_LOGOUT)
      last = thetime;
  }
  fclose(fp);
  return last;
}

int player_is_observe(int p, int g)
{
  int i;

  for (i = 0; i < player_globals.parray[p].num_observe; i++) {
    if (player_globals.parray[p].observe_list[i] == g)
      break;
  }
  if (i == player_globals.parray[p].num_observe)
    return 0;
  else
    return 1;
}

int player_add_observe(int p, int g)
{
  if (player_globals.parray[p].num_observe == MAX_OBSERVE)
    return -1;
  player_globals.parray[p].observe_list[player_globals.parray[p].num_observe] = g;
  player_globals.parray[p].num_observe++;
  return 0;
}

int player_remove_observe(int p, int g)
{
  int i;

  for (i = 0; i < player_globals.parray[p].num_observe; i++) {
    if (player_globals.parray[p].observe_list[i] == g)
      break;
  }
  if (i == player_globals.parray[p].num_observe)
    return -1;			/* Not found! */
  for (; i < player_globals.parray[p].num_observe - 1; i++) {
    player_globals.parray[p].observe_list[i] = player_globals.parray[p].observe_list[i + 1];
  }
  player_globals.parray[p].num_observe--;
  return 0;
}

int player_game_ended(int g)
{
  int p;

  for (p = 0; p < player_globals.p_num; p++) {
    if (player_globals.parray[p].status == PLAYER_EMPTY)
      continue;
    player_remove_observe(p, g);
  }
  remove_request(game_globals.garray[g].white, game_globals.garray[g].black, -1);
  remove_request(game_globals.garray[g].black, game_globals.garray[g].white, -1);
  player_save(game_globals.garray[g].white);	/* Hawk: Added to save finger-info after each
				   game */
  player_save(game_globals.garray[g].black);
  return 0;
}

int player_goto_board(int p, int board_num)
{
  int start, count = 0, on, g;

  if (player_globals.parray[p].simul_info == NULL)
    return -1;

  if (board_num < 0 || board_num >= player_globals.parray[p].simul_info->numBoards)
    return -1;
  if (player_globals.parray[p].simul_info->boards[board_num] < 0)
    return -1;
  player_globals.parray[p].simul_info->onBoard = board_num;
  player_globals.parray[p].game = player_globals.parray[p].simul_info->boards[board_num];
  player_globals.parray[p].opponent = game_globals.garray[player_globals.parray[p].game].black;
  if (player_globals.parray[p].simul_info->numBoards == 1)
    return 0;
  send_board_to(player_globals.parray[p].game, p);
  start = player_globals.parray[p].game;
  on = player_globals.parray[p].simul_info->onBoard;
  do {
    g = player_globals.parray[p].simul_info->boards[on];
    if (g >= 0) {
      if (count == 0) {
	Bell (game_globals.garray[g].black);
	pprintf(game_globals.garray[g].black, "\n");
	pprintf_highlight(game_globals.garray[g].black, "%s", player_globals.parray[p].name);
	pprintf_prompt(game_globals.garray[g].black, " is at your board!\n");
      } else if (count == 1) {
	Bell (game_globals.garray[g].black);
	pprintf(game_globals.garray[g].black, "\n");
	pprintf_highlight(game_globals.garray[g].black, "%s", player_globals.parray[p].name);
	pprintf_prompt(game_globals.garray[g].black, " will be at your board NEXT!\n");
      } else {
	pprintf(game_globals.garray[g].black, "\n");
	pprintf_highlight(game_globals.garray[g].black, "%s", player_globals.parray[p].name);
	pprintf_prompt(game_globals.garray[g].black, " is %d boards away.\n", count);
      }
      count++;
    }
    on++;
    if (on >= player_globals.parray[p].simul_info->numBoards)
      on = 0;
  } while (start != player_globals.parray[p].simul_info->boards[on]);
  return 0;
}

int player_goto_next_board(int p)
{
  int on;
  int start;
  int g;

  if (player_globals.parray[p].simul_info == NULL)
    return -1;

  on = player_globals.parray[p].simul_info->onBoard;
  start = on;
  g = -1;
  do {
    on++;
    if (on >= player_globals.parray[p].simul_info->numBoards)
      on = 0;
    g = player_globals.parray[p].simul_info->boards[on];
    if (g >= 0)
      break;
  } while (start != on);
  if (g == -1) {
    pprintf(p, "\nMajor Problem! Can't find your next board.\n");
    return -1;
  }
  return player_goto_board(p, on);
}

int player_goto_prev_board(int p)
{
  int on;
  int start;
  int g;

  if (player_globals.parray[p].simul_info == NULL)
    return -1;

  on = player_globals.parray[p].simul_info->onBoard;
  start = on;
  g = -1;
  do {
    --on;
    if (on < 0)
      on = (player_globals.parray[p].simul_info->numBoards) - 1;
    g = player_globals.parray[p].simul_info->boards[on];
    if (g >= 0)
      break;
  } while (start != on);
  if (g == -1) {
    pprintf(p, "\nMajor Problem! Can't find your previous board.\n");
    return -1;
  }
  return player_goto_board(p, on);
}

int player_goto_simulgame_bynum(int p, int num)
{
  int on;
  int start;
  int g;

  if (player_globals.parray[p].simul_info == NULL)
    return -1;

  on = player_globals.parray[p].simul_info->onBoard;
  start = on;
  do {
    on++;
    if (on >= player_globals.parray[p].simul_info->numBoards)
      on = 0;
    g = player_globals.parray[p].simul_info->boards[on];
    if (g == num)
      break;
  } while (start != on);
  if (g != num) {
    pprintf(p, "\nYou aren't playing that game!!\n");
    return -1;
  }
  return player_goto_board(p, on);
}

int player_num_active_boards(int p)
{
  int count = 0, i;

  if (player_globals.parray[p].simul_info == NULL)
    return 0;

  if (!player_globals.parray[p].simul_info->numBoards)
    return 0;
  for (i = 0; i < player_globals.parray[p].simul_info->numBoards; i++)
    if (player_globals.parray[p].simul_info->boards[i] >= 0)
      count++;
  return count;
}

static int player_num_results(int p, int result)
{
  switch (result) {
  case RESULT_WIN:
    return player_globals.parray[p].simul_info->num_wins;
  case RESULT_DRAW:
    return player_globals.parray[p].simul_info->num_draws;
  case RESULT_LOSS:
    return player_globals.parray[p].simul_info->num_losses;
  default:
    return -1;
  }
}

static void new_simul_result(simul_info_t *sim, int result)
{
  switch (result) {
  case RESULT_WIN:
    sim->num_wins++;
    return;
  case RESULT_DRAW:
    sim->num_draws++;
    return;
  case RESULT_LOSS:
    sim->num_losses++;
    return;
  }
}

int player_simul_over(int p, int g, int result)
{
  int on, ong, p1, which;
  char tmp[1024];

  if (player_globals.parray[p].simul_info == NULL)
    return -1;

  for (which = 0; which < player_globals.parray[p].simul_info->numBoards; which++) {
    if (player_globals.parray[p].simul_info->boards[which] == g) {
      break;
    }
  }
  if (which == player_globals.parray[p].simul_info->numBoards) {
    pprintf(p, "I can't find that game!\n");
    return -1;
  }
  pprintf(p, "\nBoard %d has completed.\n", which + 1);
  on = player_globals.parray[p].simul_info->onBoard;
  ong = player_globals.parray[p].simul_info->boards[on];
  player_globals.parray[p].simul_info->boards[which] = -1;
  new_simul_result(player_globals.parray[p].simul_info, result);
  if (player_num_active_boards(p) == 0) {
    sprintf(tmp, "\n{Simul (%s vs. %d) is over.}\nResults: %d Wins, %d Losses, %d Draws, %d Aborts\n",
	    player_globals.parray[p].name,
	    player_globals.parray[p].simul_info->numBoards,
	    player_num_results(p, RESULT_WIN),
	    player_num_results(p, RESULT_LOSS),
	    player_num_results(p, RESULT_DRAW),
	    player_num_results(p, RESULT_ABORT));
    for (p1 = 0; p1 < player_globals.p_num; p1++) {
      if (player_globals.parray[p].status != PLAYER_PROMPT)
	continue;
      if (!CheckPFlag(p1, PFLAG_GIN) && !player_is_observe(p1, g) && (p1 != p))
	continue;
      pprintf_prompt(p1, "%s", tmp);
    }
    player_globals.parray[p].simul_info->numBoards = 0;
    pprintf_prompt(p, "\nThat was the last board, thanks for playing.\n");
    free(player_globals.parray[p].simul_info);
    player_globals.parray[p].simul_info = NULL;
    return 0;
  }
  if (ong == g) {		/* This game is over */
    player_goto_next_board(p);
  } else {
    player_goto_board(p, player_globals.parray[p].simul_info->onBoard);
  }
  pprintf_prompt(p, "\nThere are %d boards left.\n",
		 player_num_active_boards(p));
  return 0;
}

static void GetMsgFile (int p, char *fName)
{
 sprintf(fName, "%s/player_data/%c/%s.%s", STATS_DIR, player_globals.parray[p].login[0],
         player_globals.parray[p].login, STATS_MESSAGES);
 }

int player_num_messages(int p)
{
  char fname[MAX_FILENAME_SIZE];

  if (!CheckPFlag(p, PFLAG_REG))
    return 0;
  GetMsgFile (p, fname);
  return lines_file(fname);
}

int player_add_message(int top, int fromp, char *message)
{
  char fname[MAX_FILENAME_SIZE];
  FILE *fp;
  char subj[256];
  char messbody[1024];
  time_t t = time(0);

  if (!CheckPFlag(top, PFLAG_REG))
    return -1;
  if (!CheckPFlag(fromp, PFLAG_REG))
    return -1;
  GetMsgFile (top, fname);
  if ((lines_file(fname) >= MAX_MESSAGES) && (player_globals.parray[top].adminLevel == 0))
    return -1;
  fp = fopen(fname, "a");
  if (!fp)
    return -1;
  fprintf(fp, "%s at %s: %s\n", player_globals.parray[fromp].name, strltime(&t), message);
  fclose(fp);
  pprintf(fromp, "\nThe following message was sent ");
  if (CheckPFlag(top, PFLAG_MAILMESS)) {
      sprintf(subj, "FICS message from %s at FICS %s (Do not reply by mail)", 
	      player_globals.parray[fromp].name, 
	      config_get_tmp("SERVER_HOSTNAME"));
      sprintf(messbody, "%s at %s: %s\n", player_globals.parray[fromp].name, strltime(&t), message);
      mail_string_to_user(top, subj, messbody);
      pprintf(fromp, "(and emailed) ");
  }
  pprintf(fromp, "to %s: \n   %s\n", player_globals.parray[top].name, message);
  return 0;
}

static int player_forward_message(int top, int fromp, char *message)
{
  char fname[MAX_FILENAME_SIZE];
  FILE *fp;
  char subj[256];
  char messbody[1024];

  if (!CheckPFlag(top, PFLAG_REG))
    return -1;
  if (!CheckPFlag(fromp, PFLAG_REG))
    return -1;
  GetMsgFile (top, fname);
  if ((lines_file(fname) >= MAX_MESSAGES) && (player_globals.parray[top].adminLevel == 0))
    return -1;
  fp = fopen(fname, "a");
  if (!fp)
    return -1;
  fprintf(fp, "FORWARDED MSG FROM %s : %s", player_globals.parray[fromp].name, message);
  fclose(fp);
  pprintf(fromp, "\nThe following message was forwarded ");
  if (CheckPFlag(top, PFLAG_MAILMESS)) {
      sprintf(subj, "FICS message from %s at FICS %s (Do not reply by mail)", 
	      player_globals.parray[fromp].name, 
	      config_get_tmp("SERVER_HOSTNAME"));
      sprintf(messbody, "FORWARDED MSG FROM %s: %s\n", player_globals.parray[fromp].name, message);
      mail_string_to_user(top, subj, messbody);
      pprintf(fromp, "(and emailed) ");
  }
  pprintf(fromp, "to %s: \n   %s\n", player_globals.parray[top].name, message);
  return 0;
}
void SaveTextListEntry(textlist **Entry, char *string, int n)
{
  *Entry = (textlist *) malloc(sizeof(textlist));
  (*Entry)->text = strdup(string);
  (*Entry)->index = n;
  (*Entry)->next = NULL;
}

static textlist *ClearTextListEntry(textlist *entry)
{
  textlist *ret = entry->next;
  free(entry->text);
  free(entry);
  return ret;
}

void ClearTextList(textlist *head)
{
  textlist *cur;

  for (cur = head; cur != NULL; cur = ClearTextListEntry(cur));
}

static int SaveThisMsg (int which, char *line)
{
  char Sender[MAX_LOGIN_NAME];
  int p1;

  if (which == 0) return 1;

  sscanf (line, "%s", Sender);
  if (which < 0) {
    p1 = -which - 1;
    return strcmp(Sender, player_globals.parray[p1].name);
  }
  else {
    p1 = which - 1;
    return !strcmp(Sender, player_globals.parray[p1].name);
  }
}

static int LoadMsgs(int p, int which, textlist **Head)
{
  FILE *fp;
  textlist **Cur = Head;
  char fName[MAX_FILENAME_SIZE];
  char line[MAX_LINE_SIZE];
  int n=0, nSave=0;

  *Head = NULL;
  GetMsgFile (p, fName);
  fp = fopen(fName, "r");
  if (fp == NULL) {
    return -1;
  }
  while (!feof(fp)) {
    fgets(line, MAX_LINE_SIZE, fp);
    if (feof(fp))
      break;
    if (SaveThisMsg(which, line)) {
      SaveTextListEntry(Cur, line, ++n);
      Cur = &(*Cur)->next;
      nSave++;
    }
    else n++;
  }
  fclose (fp);
  return nSave;
}

/* start > 0 and end > start (or end = 0) to save messages in range;
   start < 0 and end < start (or end = 0) to clear messages in range;
   if end = 0, range goes to end of file (not tested yet). */
static int LoadMsgRange(int p, int start, int end, textlist **Head)
{
  FILE *fp;
  char fName[MAX_FILENAME_SIZE];
  char line[MAX_LINE_SIZE];
  textlist **Cur = Head;
  int n=1, nSave=0, nKill=0;

  *Head = NULL;
  GetMsgFile (p, fName);
  fp = fopen  (fName, "r");
  if (fp == NULL) {
    pprintf (p, "You have no messages.\n");
    return -1;
  }
  for (n=1; n <= end || end <= 0; n++) {
    fgets (line, MAX_LINE_SIZE, fp);
    if (feof(fp))
      break;
    if ((start < 0 && (n < -start || n > -end)) || (start >= 0 && n >= start)) {
      SaveTextListEntry (Cur, line, n);
      Cur = &(*Cur)->next;
      nSave++;
    }
    else nKill++;
  }
  fclose (fp);
  if (start < 0) {
    if (n <= -start)
      pprintf (p, "You do not have a message %d.\n", -start);
    return nKill;
  } else {
    if (n <= start)
      pprintf (p, "You do not have a message %d.\n", start);
    return nSave;
  }
}

int ForwardMsgRange(char *p1, int p, int start, int end)
{
  /* p1 is player to   and p is player from */
  FILE *fp;
  char fName[MAX_FILENAME_SIZE];
  char line[MAX_LINE_SIZE];
  int n=1, top, connected;

  if (!FindPlayer(p, p1, &top, &connected))
    return -1;
  GetMsgFile (p, fName);
  fp = fopen  (fName, "r");
  if (fp == NULL) {
    pprintf (p, "You have no messages.\n");
    return -1;
  }
  for (n=1; n <= end || end <= 0; n++) {
    fgets (line, MAX_LINE_SIZE, fp);
    if (feof(fp))
      break;
    if ((start < 0 && (n < -start || n > -end)) || (start >= 0 && n >= start)) {
      player_forward_message(top, p, line);
    }
  }
  fclose (fp);
  if (start < 0) {
    if (n <= -start)
      pprintf (p, "You do not have a message %d.\n", -start);
    return 0;
  } else {
    if (n <= start)
      pprintf (p, "You do not have a message %d.\n", start);
    return 0;
  }
}

static int WriteMsgFile (int p, textlist *Head)
{
  char fName[MAX_FILENAME_SIZE];
  FILE *fp;
  textlist *Cur;

  GetMsgFile (p, fName);
  fp = fopen(fName, "w");
  if (fp == NULL)
    return 0;
  for (Cur = Head; Cur != NULL; Cur = Cur->next)
    fprintf(fp, "%s", Cur->text);
  fclose(fp);
  return 1;
}

int ClearMsgsBySender(int p, param_list param)
{
  textlist *Head;
  int p1, connected;
  int nFound;

  if (!FindPlayer(p, param[0].val.word, &p1, &connected))
    return -1;

  nFound = LoadMsgs(p, -(p1+1), &Head);
  if (nFound < 0) {
    pprintf(p, "You have no messages.\n");
  } else if (nFound == 0) {
    pprintf(p, "You have no messages from %s.\n", player_globals.parray[p1].name);
  } else {
    if (WriteMsgFile (p, Head))
      pprintf(p, "Messages from %s cleared.\n", player_globals.parray[p1].name); 
    else {
      pprintf(p, "Problem writing message file; please contact an admin.\n");
      fprintf(stderr, "Problem writing message file for %s.\n", player_globals.parray[p].name);
    }
    ClearTextList(Head);
  }
  if (!connected)
    player_remove(p1);
  return nFound;
}

static void ShowTextList (int p, textlist *Head, int ShowIndex)
{
  textlist *CurMsg;

  if (ShowIndex) {
    for (CurMsg = Head; CurMsg != NULL; CurMsg = CurMsg->next)
      pprintf(p, "%2d. %s", CurMsg->index, CurMsg->text);
  }
  else {
    for (CurMsg = Head; CurMsg != NULL; CurMsg = CurMsg->next)
      pprintf(p, "%s", CurMsg->text);
  }
}

int player_show_messages(int p)
{
  textlist *Head;
  int n;

  n = LoadMsgs (p, 0, &Head);
  if (n <= 0) {
    pprintf (p, "You have no messages.\n");
    return -1;
  } else {
    pprintf (p, "Messages:\n");
    ShowTextList (p, Head, 1);
    ClearTextList (Head);
    return 0;
  }
}

int ShowMsgsBySender(int p, param_list param)
{
  textlist *Head;
  int p1, connected;
  int nFrom;
  int nTo = 0;

  if (!FindPlayer(p, param[0].val.word, &p1, &connected))
    return -1;

  if (!CheckPFlag(p1, PFLAG_REG)) {
    pprintf(p, "Player \"%s\" is unregistered and cannot send or receive messages.\n",
       player_globals.parray[p1].name);
    return -1; /* no need to disconnect */
  }
 
  if (p != p1) {
    nTo = LoadMsgs(p1, p+1, &Head);
    if (nTo <= 0) {
      pprintf(p, "%s has no messages from you.\n", player_globals.parray[p1].name);
    } else {
      pprintf(p, "Messages to %s:\n", player_globals.parray[p1].name);
      ShowTextList (p, Head, 0);
      ClearTextList(Head);
    }
  }
  nFrom = LoadMsgs(p, p1+1, &Head);
  if (nFrom <= 0) {
    pprintf(p, "\nYou have no messages from %s.\n", player_globals.parray[p1].name);
  } else {
    pprintf(p, "Messages from %s:\n", player_globals.parray[p1].name);
    ShowTextList (p, Head, 1);
    ClearTextList(Head);
  }
  if (!connected)
    player_remove(p1);
  return (nFrom > 0 || nTo > 0);
}

int ShowMsgRange (int p, int start, int end)
{
  textlist *Head;
  int n;

  n = LoadMsgRange (p, start, end, &Head);
  if (n > 0) {
    ShowTextList (p, Head, 1);
    ClearTextList (Head);
  }
  return n;
}

int ClrMsgRange (int p, int start, int end)
{
  textlist *Head;
  int n;

  n = LoadMsgRange (p, -start, -end, &Head);
  if (n > 0)
    if (WriteMsgFile (p, Head))
      pprintf (p, "Message %d cleared.\n", start);
  if (n >= 0)
    ClearTextList (Head);

  return n;
}

int player_clear_messages(int p)
{
  char fname[MAX_FILENAME_SIZE];

  if (!CheckPFlag(p, PFLAG_REG))
    return -1;
  GetMsgFile (p, fname);
  unlink(fname);
  return 0;
}

int player_search(int p, char *name)
/*
 * Find player matching the given string. First looks for exact match
 *  with a logged in player, then an exact match with a registered player,
 *  then a partial unique match with a logged in player, then a partial
 *  match with a registered player.
 *  Returns player number if the player is connected, negative (player number)
 *  if the player had to be connected, and 0 if no player was found
 */
{
  int p1, count;
  char *buffer[1000];
  char pdir[MAX_FILENAME_SIZE];

  /* exact match with connected player? */
  if ((p1 = player_find_bylogin(name)) >= 0) {
    return p1 + 1;
  }
  /* exact match with registered player? */
  sprintf(pdir, "%s/%c", PLAYER_DIR, name[0]);
  count = search_directory(pdir, name, buffer, 1000);
  if (count > 0 && !strcmp(name, *buffer)) {
    goto ReadPlayerFromFile;	/* found an unconnected registered player */
  }
  /* partial match with connected player? */
  if ((p1 = player_find_part_login(name)) >= 0) {
    return p1 + 1;
  } else if (p1 == -2) {
    /* ambiguous; matches too many connected players. */
    pprintf (p, "Ambiguous name '%s'; matches more than one player.\n", name);
    return 0;
  }
  /* partial match with registered player? */
  if (count < 1) {
    pprintf(p, "There is no player matching that name.\n");
    return 0;
  }
  if (count > 1) {
    pprintf(p, "-- Matches: %d names --", count);
    display_directory(p, buffer, count);
    return(0);
/*
    char *s = buffer;
    multicol *m = multicol_start(2000);
    for (i = 0; i < count; i++) {
      multicol_store(m, s);
      s += strlen(s) + 1;
    }
    multicol_pprint(m, p, player_globals.parray[p].d_width, 1);
    multicol_end(m);
    return 0;
*/
  }
ReadPlayerFromFile:
  p1 = player_new();
  if (player_read(p1, *buffer)) {
    player_remove(p1);
    pprintf(p, "ERROR: a player named %s was expected but not found!\n",
	    *buffer);
    pprintf(p, "Please tell an admin about this incident. Thank you.\n");
    return 0;
  }
  return (-p1) - 1;		/* negative to indicate player was not
				   connected */
}


int player_kill(char *name)
{
  char fname[MAX_FILENAME_SIZE], fname2[MAX_FILENAME_SIZE];

  sprintf(fname, "%s/%c/%s", PLAYER_DIR, name[0], name);
  sprintf(fname2, "%s/%c/.rem.%s", PLAYER_DIR, name[0], name);
  rename(fname, fname2);
  RemHist (name);
  sprintf(fname, "%s/player_data/%c/%s.games", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.games", STATS_DIR, name[0], name);
  rename(fname, fname2);
  sprintf(fname, "%s/player_data/%c/%s.comments", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.comments", STATS_DIR, name[0], name);
  rename(fname, fname2);

  sprintf(fname, "%s/player_data/%c/%s.logons", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.logons", STATS_DIR, name[0], name);
  rename(fname, fname2);
  sprintf(fname, "%s/player_data/%c/%s.messages", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.messages", STATS_DIR, name[0], name);
  rename(fname, fname2);
  return 0;
}

int player_rename(char *name, char *newname)
{
  char fname[MAX_FILENAME_SIZE], fname2[MAX_FILENAME_SIZE];

  sprintf(fname, "%s/%c/%s", PLAYER_DIR, name[0], name);
  sprintf(fname2, "%s/%c/%s", PLAYER_DIR, newname[0], newname);
  rename(fname, fname2);
  sprintf(fname, "%s/player_data/%c/%s.games", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/%s.games", STATS_DIR, newname[0], newname);
  rename(fname, fname2);
  sprintf(fname, "%s/player_data/%c/%s.comments", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/%s.comments", STATS_DIR, newname[0], newname);
  rename(fname, fname2);
  sprintf(fname, "%s/player_data/%c/%s.logons", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/%s.logons", STATS_DIR, newname[0], newname);
  rename(fname, fname2);
  sprintf(fname, "%s/player_data/%c/%s.messages", STATS_DIR, name[0], name);
  sprintf(fname2, "%s/player_data/%c/%s.messages", STATS_DIR, newname[0], newname);
  rename(fname, fname2);
  return 0;
}

int player_reincarn(char *name, char *newname)
{
  char fname[MAX_FILENAME_SIZE], fname2[MAX_FILENAME_SIZE];

  sprintf(fname, "%s/%c/%s", PLAYER_DIR, newname[0], newname);
  sprintf(fname2, "%s/%c/.rem.%s", PLAYER_DIR, name[0], name);
  rename(fname2, fname);
  sprintf(fname, "%s/player_data/%c/%s.games", STATS_DIR, newname[0], newname);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.games", STATS_DIR, name[0], name);
  rename(fname2, fname);
  sprintf(fname, "%s/player_data/%c/%s.comments", STATS_DIR, newname[0], newname);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.comments", STATS_DIR, name[0], name);
  rename(fname2, fname);
  sprintf(fname, "%s/player_data/%c/%s.logons", STATS_DIR, newname[0], newname);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.logons", STATS_DIR, name[0], name);
  rename(fname2, fname);
  sprintf(fname, "%s/player_data/%c/%s.messages", STATS_DIR, newname[0], newname);
  sprintf(fname2, "%s/player_data/%c/.rem.%s.messages", STATS_DIR, name[0], name);
  rename(fname2, fname);
  return 0;
}

int player_num_comments(int p)
{
  char fname[MAX_FILENAME_SIZE];

  if (!CheckPFlag(p, PFLAG_REG))
    return 0;
  sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR, player_globals.parray[p].login[0],
	  player_globals.parray[p].login, "comments");
  return lines_file(fname);
}

int player_add_comment(int p_by, int p_to, char *comment)
{
  char fname[MAX_FILENAME_SIZE];
  FILE *fp;
  time_t t = time(0);

  if (!CheckPFlag(p_to, PFLAG_REG))
    return -1;
  sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR, player_globals.parray[p_to].login[0],
	  player_globals.parray[p_to].login, "comments");
  fp = fopen(fname, "a");
  if (!fp)
    return -1;
  fprintf(fp, "%s at %s: %s\n", player_globals.parray[p_by].name, strltime(&t), comment);
  fclose(fp);
  player_globals.parray[p_to].num_comments = player_num_comments(p_to);
  return 0;
}

int player_show_comments(int p, int p1)
{
  char fname[MAX_FILENAME_SIZE];

  if (CheckPFlag(p1, PFLAG_REG)) {
    sprintf(fname, "%s/player_data/%c/%s.%s", STATS_DIR, player_globals.parray[p1].login[0],
  	  player_globals.parray[p1].login, "comments");
    if (psend_file(p, NULL, fname))
      pprintf(p, "There are no comments to show for %s.\n", player_globals.parray[p1].name);
    if (player_globals.parray[p1].passwd[0] == '*')
      pprintf(p, "%s's account is LOCKED.\n", player_globals.parray[p1].name);
    if (in_list(p1, L_BAN, player_globals.parray[p1].name))
      pprintf(p, "%s has been BANNED.\n", player_globals.parray[p1].name);
  } else
    pprintf(p, "Player \"%s\" is unregistered and cannot have comments.\n",
               player_globals.parray[p1].name);
  return 0;
}

/* returns 1 if player is head admin, 0 otherwise */
int player_ishead(int p)
{
  return (strcasecmp(player_globals.parray[p].name, config_get_tmp("HEAD_ADMIN")) == 0);
}

/* GetRating chooses between blitz, standard and other ratings. */
int GetRating(player *p, int gametype)
{
    if (gametype == TYPE_BLITZ) return (p->b_stats.rating);
    else if (gametype == TYPE_STAND) return (p->s_stats.rating);
    else if (gametype == TYPE_WILD) return (p->w_stats.rating);
    else if (gametype == TYPE_LIGHT) return (p->l_stats.rating);
    else if (gametype == TYPE_BUGHOUSE) return (p->bug_stats.rating);
    else return 0;
}    /* end of function GetRating. */

/* GetRD chooses between blitz, standard and other RD's. */
double GetRD(player *p, int gametype)
{
  statistics *s;

  switch(gametype) {
    case TYPE_BLITZ:   s = &p->b_stats;   break;
    case TYPE_STAND:   s = &p->s_stats;   break;
    case TYPE_WILD:   s = &p->w_stats;   break;
    case TYPE_LIGHT:   s = &p->l_stats;   break;
    case TYPE_BUGHOUSE:   s = &p->bug_stats;   break;
    default:   return 0.0;
  }
  return (current_sterr(s->sterr, time(NULL)-(s->ltime)));
}    /* end of function GetRD. */

